############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

import bottle
import json
import os
import tarfile
import itertools
import operator

from accelerator.job import Job
from accelerator.dataset import Dataset

def get_job(jobid):
	if jobid.endswith('-LATEST'):
		base = jobid.rsplit('-', 1)[0]
		jobid = os.readlink(Job(base + '-0').path[:-2] + '-LATEST')
	return Job(jobid)

def main(argv, cfg):
	prog = argv.pop(0)
	if '-h' in argv or '--help' in argv or len(argv) not in (0, 1):
		print('usage: %s [port]' % (prog,))
		print('runs a web server on port (default 8520)')
		print('for displaying results (result_directory)')
		return
	port = int(argv[0]) if argv else 8520

	@bottle.get('/')
	@bottle.view('main')
	def main_page():
		return dict(
			project=os.path.split(cfg.project_directory)[1],
			workdirs=cfg.workdirs,
		)

	@bottle.get('/results')
	def results():
		res = {}
		for fn in os.listdir(cfg.result_directory):
			if fn.endswith('_'):
				continue
			ffn = os.path.join(cfg.result_directory, fn)
			try:
				jobid = os.readlink(ffn).split('/')[-2]
				res[fn] = dict(
					jobid=jobid,
					ts=os.lstat(ffn).st_mtime,
					size=os.stat(ffn).st_size,
				)
			except OSError:
				continue
		bottle.response.content_type = 'application/json; charset=UTF-8'
		bottle.response.set_header('Cache-Control', 'no-cache')
		return json.dumps(res)

	@bottle.get('/results/<name>')
	def file(name):
		return bottle.static_file(name, root=cfg.result_directory)

	@bottle.get('/job/<jobid>/method.tar.gz/')
	@bottle.get('/job/<jobid>/method.tar.gz/<name:path>')
	def job_method(jobid, name=None):
		job = get_job(jobid)
		with tarfile.open(job.filename('method.tar.gz'), 'r:gz') as tar:
			if name:
				info = tar.getmember(name)
			else:
				members = [info for info in tar.getmembers() if info.isfile()]
				if len(members) == 1 and not name:
					info = members[0]
				else:
					return bottle.template('job_method_list', members=members, job=job)
			bottle.response.content_type = 'text/plain; charset=UTF-8'
			return tar.extractfile(info).read()

	@bottle.get('/job/<jobid>/<name:path>')
	def job_file(jobid, name):
		job = get_job(jobid)
		return bottle.static_file(name, root=job.path)

	@bottle.get('/job/<jobid>')
	@bottle.get('/job/<jobid>/')
	@bottle.view('job')
	def job(jobid):
		job = get_job(jobid)
		try:
			prefix = job.path + '/'
			files = [fn[len(prefix):] for fn in job.files() if fn.startswith(prefix)]
		except OSError:
			# This happens for jobs that didn't finish.
			files = []
		subjobs = [Job(jobid) for jobid in job.post.subjobs]
		return dict(
			job=job,
			output=os.path.exists(job.filename('OUTPUT')),
			datasets=job.datasets,
			params=job.params,
			subjobs=subjobs,
			files=files,
		)

	@bottle.get('/dataset/<dsid:path>')
	def dataset(dsid):
		ds = Dataset(dsid.rstrip('/'))
		q = bottle.request.query
		if q.column:
			lines = int(q.lines or 10)
			it = ds.iterate(None, q.column)
			it = itertools.islice(it, lines)
			t = ds.columns[q.column].type
			if t in ('datetime', 'date', 'time',):
				it = map(str, it)
			elif t in ('bytes', 'pickle',):
				it = map(repr, it)
			res = list(it)
			bottle.response.content_type = 'application/json; charset=UTF-8'
			return json.dumps(res)
		else:
			return bottle.template('dataset', ds=ds)

	@bottle.get('/workdir/<name>')
	@bottle.view('workdir')
	def workdir(name):
		path = cfg.workdirs[name]
		prefix = name + '-'
		jobs = []
		for jid in os.listdir(path):
			if jid.startswith(prefix):
				try:
					jobs.append(Job(jid))
				except Exception:
					pass
		jobs.sort(key=operator.attrgetter('number'))
		return dict(name=name, jobs=jobs)

	bottle.TEMPLATE_PATH = [os.path.join(os.path.dirname(__file__), 'board')]
	bottle.run(port=port, reloader=True)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019 Carl Drougge                            #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# workarounds for sadnesses

import os
import signal
import termios
import fcntl
import atexit
from contextlib import contextmanager

class SignalWrapper(object):
	"""Some misguided kernels (like Linux) feel that SIGINFO is not needed.
	Perhapt some other kernels feel similarly about other signals?"""

	def __init__(self, signal_names=['SIGINFO'], key_values=[20], skip_input_if_possible=True):
		"""signal_names is a list of signal names, which will be listened for if they exist.
		key_values is a list of (terminal input) key values, each will
		be considered equivalent of a signal if pressed.
		If skip_input_if_possible is True no input will be read if all signals exist.
		No differentiation is made between the signals."""

		self.key_values = set(key_values)
		self.restore = {}
		self.signal_set = False
		self.clean = False
		self.use_input = False
		all_sigs = True
		atexit.register(self.cleanup)
		for name in signal_names:
			sig = getattr(signal, name, None)
			if sig is not None:
				old = signal.signal(sig, self.signal_arrived)
				self.restore[sig] = old
				signal.siginterrupt(sig, False)
			else:
				all_sigs = False
		if all_sigs and skip_input_if_possible:
			return
		self.tc_original = termios.tcgetattr(0)
		tc_changed = list(self.tc_original)
		tc_changed[3] &= ~(termios.ICANON | termios.IEXTEN)
		self.use_input = True
		termios.tcsetattr(0, termios.TCSADRAIN, tc_changed)

	def cleanup(self):
		if not self.clean:
			for sig, old in self.restore.items():
				if old is None:
					old = signal.SIG_DFL
				signal.signal(sig, old)
			if self.use_input:
				termios.tcsetattr(0, termios.TCSADRAIN, self.tc_original)
			self.clean = True

	def check(self):
		if self.use_input:
			with nonblocking(0):
				while True:
					try:
						if ord(os.read(0, 1)) in self.key_values:
							self.signal_set = True
					except OSError:
						break
		if self.signal_set:
			self.signal_set = False
			return True
		else:
			return False

	def signal_arrived(self, sig, frame):
		self.signal_set = True

@contextmanager
def nonblocking(fd):
	fl_original = fcntl.fcntl(fd, fcntl.F_GETFL)
	fl_changed = fl_original | os.O_NONBLOCK
	fcntl.fcntl(fd, fcntl.F_SETFL, fl_changed)
	try:
		yield
	finally:
		fcntl.fcntl(fd, fcntl.F_SETFL, fl_original)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019 Carl Drougge                            #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

from traceback import print_exc
from datetime import datetime, date, time, timedelta

from accelerator.compat import iteritems, itervalues, first_value, str_types, int_types, num_types, unicode

from accelerator.extras import OptionEnum, OptionEnumValue, OptionString, OptionDefault, RequiredOption, typing_conv
from accelerator.job import JobWithFile

class OptionException(Exception):
	pass

_date_types = (datetime, date, time, timedelta)

class DepTree:

	def __init__(self, methods, setup):
		tree = methods.new_deptree(setup.method)
		self.methods = methods
		self.top_method = setup.method
		self.tree = tree
		self.add_flags({'make' : False, 'link' : False, })
		seen = set()
		for method, data in iteritems(self.tree):
			seen.add(method)
			data['params'] = {method: setup.params[method]}
		unmatched = {method: params for method, params in iteritems(setup.params) if method not in seen}
		if unmatched:
			from accelerator.extras import json_encode
			print("DepTree Warning:  Unmatched options remain:", json_encode(unmatched, as_str=True))
		def collect(method):
			# All methods that method depend on
			for child in tree[method]['dep']:
				yield child
				for method in collect(child):
					yield method
		# This probably updates some with the same data several times,
		# but this is cheap (key: dictref updates, nothing more.)
		for method, data in iteritems(self.tree):
			for submethod in set(collect(method)):
				data['params'].update(tree[submethod]['params'])
		self._fix_options(False)
		self._fix_jobids('jobs')
		self._fix_jobids('datasets')

	def add_flags(self, flags):
		uid = 0
		for x, y in self.tree.items():
			y.update(flags)
			y.update({'uid' : uid, })
			uid += 1

	def get_reqlist(self):
		for method, data in self.tree.items():
			full_params = {}
			for submethod, given_params in iteritems(data['params']):
				params = {k: dict(v) for k, v in iteritems(self.methods.params[submethod].defaults)}
				for k, v in iteritems(given_params):
					params[k].update(v)
				full_params[submethod] = params
			yield method, data['uid'], self.methods.params2optset(full_params)

	def fill_in_default_options(self):
		self._fix_options(True)

	def _fix_jobids(self, key):
		for method, data in iteritems(self.tree):
			method_params = data['params'][method]
			data = method_params[key]
			method_wants = self.methods.params[method][key]
			res = {}
			for jobid_name in method_wants:
				if isinstance(jobid_name, str_types):
					value = data.get(jobid_name)
					assert value is None or isinstance(value, str), 'Input %s on %s not a string as required' % (jobid_name, method,)
				elif isinstance(jobid_name, list):
					if len(jobid_name) != 1 or not isinstance(jobid_name[0], str_types):
						raise OptionException('Bad %s item on %s: %s' % (key, method, repr(jobid_name),))
					jobid_name = jobid_name[0]
					value = data.get(jobid_name)
					if value:
						if isinstance(value, str_types):
							value = [e.strip() for e in value.split(',')]
					else:
						value = []
					assert isinstance(value, list), 'Input %s on %s not a list or string as required' % (jobid_name, method,)
				else:
					raise OptionException('%s item of unknown type %s on %s: %s' % (key, type(jobid_name), method, repr(jobid_name),))
				res[jobid_name] = value
			method_params[key] = res
			spill = set(data) - set(res)
			if spill:
				raise OptionException('Unknown %s on %s: %s' % (key, method, ', '.join(sorted(spill)),))

	def _fix_options(self, fill_in):
		for method, data in iteritems(self.tree):
			data = data['params'][method]
			options = self.methods.params[method].options
			res_options = {}
			def typefuzz(t):
				if issubclass(t, str_types):
					return str_types
				if issubclass(t, int_types):
					return int_types
				return t
			def convert(default_v, v):
				if isinstance(default_v, RequiredOption):
					if v is None and not default_v.none_ok:
						raise OptionException('Option %s on method %s requires a non-None value (%r)' % (k, method, default_v.value,))
					default_v = default_v.value
				if default_v is None or v is None:
					if default_v is OptionString:
						raise OptionException('Option %s on method %s requires a non-empty string value' % (k, method,))
					if hasattr(default_v, '_valid') and v not in default_v._valid:
						raise OptionException('Option %s on method %s requires a value in %s' % (k, method, default_v._valid,))
					if isinstance(default_v, OptionDefault):
						v = default_v.default
					return v
				if isinstance(default_v, OptionDefault):
					default_v = default_v.value
				if isinstance(default_v, dict) and isinstance(v, dict):
					if default_v:
						sample_v = first_value(default_v)
						for chk_v in itervalues(default_v):
							assert isinstance(chk_v, type(sample_v))
						return {k: convert(sample_v, v) for k, v in iteritems(v)}
					else:
						return v
				if isinstance(default_v, (list, set, tuple,)) and isinstance(v, str_types + (list, set, tuple,)):
					if isinstance(v, str_types):
						v = (e.strip() for e in v.split(','))
					if default_v:
						sample_v = first_value(default_v)
						for chk_v in default_v:
							assert isinstance(chk_v, type(sample_v))
						v = (convert(sample_v, e) for e in v)
					return type(default_v)(v)
				if isinstance(default_v, (OptionEnum, OptionEnumValue,)):
					if not (v or None) in default_v._valid:
						ok = False
						for cand_prefix in default_v._prefixes:
							if v.startswith(cand_prefix):
								ok = True
								break
						if not ok:
							raise OptionException('%r not a permitted value for option %s on method %s (%s)' % (v, k, method, default_v._valid))
					return v or None
				if isinstance(default_v, str_types + num_types) and isinstance(v, str_types + num_types):
					if default_v is OptionString:
						v = str(v)
						if not v:
							raise OptionException('Option %s on method %s requires a non-empty string value' % (k, method,))
						return v
					if isinstance(default_v, unicode) and isinstance(v, bytes):
						return v.decode('utf-8')
					return type(default_v)(v)
				if (isinstance(default_v, type) and isinstance(v, typefuzz(default_v))) or isinstance(v, typefuzz(type(default_v))):
					return v
				if isinstance(default_v, bool) and isinstance(v, (str, int)):
					lv = str(v).lower()
					if lv in ('true', '1', 't', 'yes', 'on',):
						return True
					if lv in ('false', '0', 'f', 'no', 'off', '',):
						return False
				if isinstance(default_v, _date_types):
					default_v = type(default_v)
				if default_v in _date_types:
					try:
						return typing_conv[default_v.__name__](v)
					except Exception:
						raise OptionException('Failed to convert option %s %r to %s on method %s' % (k, v, default_v, method,))
				if isinstance(v, str_types) and not v:
					return type(default_v)()
				if isinstance(default_v, JobWithFile) or default_v is JobWithFile:
					defaults = ('', '', False, None,)
					if default_v is JobWithFile:
						default_v = defaults
					if not isinstance(v, (list, tuple,)) or not (2 <= len(v) <= 4):
						raise OptionException('Option %s (%r) on method %s is not %s compatible' % (k, v, method, type(default_v)))
					v = tuple(v) + defaults[len(v):] # so all of default_v gets convert()ed.
					v = [convert(dv, vv) for dv, vv in zip(default_v, v)]
					return JobWithFile(*v)
				raise OptionException('Failed to convert option %s of %s to %s on method %s' % (k, type(v), type(default_v), method,))
			for k, v in iteritems(data['options']):
				if k in options:
					try:
						res_options[k] = convert(options[k], v)
					except OptionException:
						raise
					except Exception:
						print_exc()
						raise OptionException('Failed to convert option %s on method %s' % (k, method,))
				else:
					raise OptionException('Unknown option %s on method %s' % (k, method,))
			if fill_in:
				missing = set(options) - set(res_options)
				missing_required = missing & self.methods.params[method].required
				if missing_required:
					raise OptionException('Missing required options {%s} on method %s' % (', '.join(sorted(missing_required)), method,))
				defaults = self.methods.params[method].defaults
				res_options.update({k: defaults.options[k] for k in missing})
			data['options'] = res_options

	def get_item_by_uid(self, uid):
		for v in itervalues(self.tree):
			if v['uid'] == uid:
				return v

	def set_link(self, uid, job):
		item = self.get_item_by_uid(uid)
		item['link'] = job.id
		item['total_time'] = job.total

	def propagate_make(self):
		self._recursive_propagate_make(self.top_method)

	def _recursive_propagate_make(self, node):
		for child in self.tree[node]['dep']:
			self._recursive_propagate_make(child)
		self.tree[node]['make'] = (
			True in [self.tree[x]['make'] for x in self.tree[node]['dep'] if x]) or (
			not self.tree[node]['link'])

	def get_sorted_joblist(self):
		return sorted(self.tree.values(), key = lambda x : x['level'])

	def get_link(self, node):
		return self.tree[node]['link']


	def debugprint(self):
		for x, y in sorted(self.tree.items(), key = lambda x: -int(x[1]['level'])):
			print(' %15s' % x, end=' ')
			for k in y:
				print('%5s=%5s' % (k, y[k]), end=' ')
		print()



############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import absolute_import

description = r'''
new_ds.iterate(None) gives the same order as old_ds.iterate('roundrobin').

This is useful to get the original file order from csvimport, one
slice at a time.

If you don't specify trigger_column you get the same number of lines per
slice as old_ds. If you do specify trigger_column slice switching is
delayed until the value in trigger_column changes. This is so jobs using
new_ds get all lines with the same trigger_column value in the same
slice, useful for a number of algorithms. (Assuming all occurances of
the same value are together.)
'''

from accelerator.compat import izip

options = dict(trigger_column=None)

datasets = ('source', 'previous',)

def prepare(job):
	d = datasets.source
	dw = job.datasetwriter(
		columns=d.columns,
		caption=d.caption,
		filename=d.filename,
		previous=datasets.previous,
	)
	if options.trigger_column:
		assert options.trigger_column in datasets.source.columns, "Trigger column %r not in %s" % (options.trigger_column, datasets.source,)
		ix = sorted(datasets.source.columns).index(options.trigger_column)
	else:
		ix = -1
	return dw, ix

def analysis(sliceno, prepare_res):
	write = prepare_res[0].write_list
	ix = prepare_res[1]
	d = datasets.source
	to_copy = d.lines[sliceno]
	if to_copy == 0:
		# bail out empty slices right away
		return
	to_skip = sum(d.lines[:sliceno])
	if to_skip:
		it = d.iterate('roundrobin', slice=to_skip - bool(options.trigger_column))
		if options.trigger_column:
			trigger_v = next(it)[ix]
			# keep skipping until trigger value changes
			for v in it:
				to_copy -= 1
				if v[ix] != trigger_v:
					write(v)
					break
				if to_copy == 0:
					return # no lines left for this slice
	else:
		it = d.iterate('roundrobin')
	# write the lines belonging here
	# (zip so we don't have to count down to_copy manually)
	for _, v in izip(range(to_copy), it):
		write(v)
	if options.trigger_column:
		trigger_v = v[ix]
		# keep copying until trigger value changes or lines run out
		for v in it:
			if trigger_v != v[ix]:
				break
			write(v)

def synthesis(prepare_res):
	want = datasets.source.lines
	got = prepare_res[0].finish().lines
	if options.trigger_column:
		# Lines will have moved, but the total must still be the same.
		want = sum(want)
		got = sum(got)
	assert want == got, "Did not get the expected line counts"



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2020 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import absolute_import

description = r"""Make only some columns from a dataset visible."""

from accelerator.extras import RequiredOption

options = dict(
	columns = RequiredOption(["colname1", "colname2", "..."]),
)

datasets = ("source",)

def synthesis():
	datasets.source.link_to_here(column_filter=options.columns)



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test writing a dataset in prepare, verifying that it is usable in
analysis and synthesis with no manual .finish()
'''

from accelerator.dataset import DatasetWriter, Dataset

def prepare():
	dw = DatasetWriter(columns={"data": "ascii"})
	write = dw.get_split_write()
	write("foo")
	write("bar")

def analysis(sliceno, params):
	ds = Dataset(params.jobid)
	assert set(ds.iterate(None, "data")) == {"foo", "bar"}

def synthesis(params):
	ds = Dataset(params.jobid)
	assert set(ds.iterate(None, "data")) == {"foo", "bar"}



############################################################################
#                                                                          #
# Copyright (c) 2019-2020 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify a few corner cases in the json functions in extras.
'''

from collections import OrderedDict
from itertools import permutations

from accelerator.extras import json_save, json_load, json_encode
from accelerator.compat import PY2, PY3

def test(name, input, want_obj, want_bytes, **kw):
	json_save(input, name, **kw)
	with open(name, "rb") as fh:
		got_bytes_raw = fh.read()
		assert got_bytes_raw[-1:] == b"\n", name + " didn't even end with a newline"
		got_bytes_raw = got_bytes_raw[:-1]
	as_str = json_encode(input, as_str=True, **kw)
	as_bytes = json_encode(input, as_str=False, **kw)
	assert isinstance(as_str, str) and isinstance(as_bytes, bytes), "json_encode returns the wrong types: %s %s" % (type(as_str), type(as_bytes),)
	assert as_bytes == got_bytes_raw, "json_save doesn't save the same thing json_encode returns for " + name
	if PY3:
		as_str = as_str.encode("utf-8")
	assert as_bytes == as_str, "json_encode doesn't return the same data for as_str=True and False"
	got_obj = json_load(name)
	assert want_obj == got_obj, "%s roundtrips wrong (wanted %r, got %r)" % (name, want_obj, got_obj)
	with open(name, "rb") as fh:
		got_bytes_fuzzy = b"".join(line.strip() for line in fh)
	assert want_bytes == got_bytes_fuzzy, "%s wrong on disk (but decoded right)" % (name,)

def synthesis():
	test(
		"simple.json",
		dict(a=1, b="2"),
		dict(a=1, b="2"),
		b'{"a": 1,"b": "2"}',
	)
	test(
		"True False None.json",
		[True, False, None],
		[True, False, None],
		b'[true,false,null]',
	)
	test(
		"list like types.json",
		dict(
			list=[1, 2, 3,],
			set={"foo"},
			tuple=("a", "b", "c",),
		),
		dict(
			list=[1, 2, 3,],
			set=["foo"],
			tuple=["a", "b", "c",],
		),
		b'{"list": [1,2,3],"set": ["foo"],"tuple": ["a","b","c"]}',
	)

	unicode_want = u"bl\xe4"
	if PY2:
		unicode_want = unicode_want.encode("utf-8")
	test(
		"unicode.json",
		u"bl\xe4",
		unicode_want,
		b'"bl\\u00e4"',
	)

	# Verify that utf-8 encoding also works for reading.
	with open("utf-8.json", "wb") as fh:
		fh.write(b'"bl\xc3\xa4"')
	assert json_load("utf-8.json") == unicode_want

	# This is supposed to work on PY2, but not PY3.
	try:
		test(
			"string encoding.json",
			[u"\xe4", b"\xc3\xa4", [b"\xe4", {b"\xe4": b"\xc3\xa4",},],],
			[b"\xc3\xa4", b"\xc3\xa4", [b"\xc3\xa4", {b"\xc3\xa4": b"\xc3\xa4",},],],
			b'["\\u00e4","\\u00e4",["\\u00e4",{"\\u00e4": "\\u00e4"}]]',
		)
		string_encoding_ok = True
	except TypeError:
		string_encoding_ok = False
	if PY2:
		assert string_encoding_ok, "Bytes are supposed to work in json_encode on PY2"
	else:
		assert not string_encoding_ok, "Bytes are not supposed to work in json_encode on PY3"

	# 720 permutations might be a bit much, but at least it's unlikely to
	# miss ordering problems.
	sorted_s = None
	for ix, pairs in enumerate(permutations(zip("abcdef", range(6)))):
		d = OrderedDict()
		s = "{"
		for k, v in pairs:
			d[k] = v
			s += '"%s": %s,' % (k, v)
		s = (s[:-1] + "}").encode("ascii")
		if not sorted_s:
			sorted_s = s
		test("ordered%d.json" % (ix,), d, d, s, sort_keys=False)
		test("sorted%d.json" % (ix,), d, d, sorted_s, sort_keys=True)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2020 Anders Berkeman                    #
# Modifications copyright (c) 2018-2020 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

import time
import sys
import os
import json
from operator import itemgetter
from collections import defaultdict
from datetime import date
from base64 import b64encode
from importlib import import_module
from argparse import ArgumentParser, RawTextHelpFormatter

from accelerator.compat import unicode, str_types, PY3
from accelerator.compat import urlencode, urlopen, Request, URLError, HTTPError
from accelerator.compat import getarglist

from accelerator import setupfile
from accelerator.extras import json_encode, json_decode, DotDict, _ListTypePreserver
from accelerator.job import Job
from accelerator.status import print_status_stacks
from accelerator.error import JobError, ServerError, UrdError, UrdPermissionError, UrdConflictError
from accelerator import g
from accelerator import unixhttp; unixhttp # for unixhttp:// URLs, as used to talk to the server


class Automata:
	"""
	Launch jobs, wait for completion.
	Don't use this directly, use the Urd object.
	"""

	method = '?' # fall-through case when we resume waiting for something

	def __init__(self, server_url, verbose=False, flags=None, subjob_cookie=None, infoprints=False, print_full_jobpath=False):
		self.url = server_url
		self.subjob_cookie = subjob_cookie
		self.history = []
		self.verbose = verbose
		self.monitor = None
		self.flags = flags or []
		self.job_method = None
		# Workspaces should be per Automata
		from accelerator.job import WORKDIRS
		WORKDIRS.update(self.list_workdirs())
		self.update_method_info()
		self.clear_record()
		# Only do this when run from shell.
		if infoprints:
			from accelerator.workarounds import SignalWrapper
			siginfo = SignalWrapper(['SIGINFO', 'SIGUSR1'])
			self.siginfo_check = siginfo.check
		else:
			self.siginfo_check = lambda: False
		self.print_full_jobpath = print_full_jobpath

	def clear_record(self):
		self.record = defaultdict(JobList)
		self.jobs = self.record[None]

	def validate_response(self, response):
		# replace with homemade function,
		# this is run on bigdata response
		pass

	def _url_get(self, *path, **kw):
		url = self.url + os.path.join('/', *path)
		for attempt in (1, 2, 3, 4, 5):
			resp = None
			try:
				req = urlopen(url, **kw)
				try:
					resp = req.read()
					if PY3:
						resp = resp.decode('utf-8')
					code = req.getcode()
					if code == 503:
						# It seems inconsistent if we get HTTPError or not for these.
						raise HTTPError(url, code, '', {}, None)
					return resp
				finally:
					try:
						req.close()
					except Exception:
						pass
			except HTTPError as e:
				if e.code != 503:
					return resp
				msg = 'Server says %d: %s' % (e.code, resp,)
			except URLError:
				# Don't say anything the first times, because the output
				# tests get messed up if this happens during them.
				if attempt < 3:
					msg = None
				else:
					msg = 'Error contacting server'
			if msg:
				print(msg, file=sys.stderr)
			if attempt < 5:
				time.sleep(attempt / 15)
				if msg:
					print('Retrying (%d/4).' % (attempt,), file=sys.stderr)
		print('Giving up.', file=sys.stderr)
		raise ServerError(msg)

	def _url_json(self, *path, **kw):
		return json_decode(self._url_get(*path, **kw))

	def abort(self):
		return self._url_json('abort')

	def info(self):
		return self._url_json('workspace_info')

	def config(self):
		return self._url_json('config')

	def _submit(self, method, options, datasets, jobs, caption=None, wait=True, why_build=False, workdir=None):
		"""
		Submit job to server and conditionaly wait for completion.
		"""
		self.job_method = method
		if not why_build and 'why_build' in self.flags:
			why_build = 'on_build'
		if self.monitor and not why_build:
			self.monitor.submit(method)
		if not caption:
			caption = ''
		params = {method: dict(options=options, datasets=datasets, jobs=jobs,)}
		data = setupfile.generate(caption, method, params, why_build=why_build)
		if self.subjob_cookie:
			data.subjob_cookie = self.subjob_cookie
			data.parent_pid = os.getpid()
		if workdir:
			data.workdir = workdir
		t0 = time.time()
		self.job_retur = self._server_submit(data)
		self.history.append((data, self.job_retur))
		#
		if wait and not self.job_retur.done:
			self.wait(t0)
		if self.monitor and not why_build:
			self.monitor.done()
		return self.jobid(method), self.job_retur

	def wait(self, t0=None, ignore_old_errors=False):
		idle, status_stacks, current, last_time = self._server_idle(0, ignore_errors=ignore_old_errors)
		if idle:
			return
		if t0 is None:
			if current:
				t0 = current[0]
			else:
				t0 = time.time()
		waited = int(round(time.time() - t0)) - 1
		if self.verbose == 'dots':
			print('[' + '.' * waited, end=' ')
		while not idle:
			if self.siginfo_check():
				print()
				print_status_stacks(status_stacks)
			waited += 1
			if waited % 60 == 0 and self.monitor:
				self.monitor.ping()
			if self.verbose:
				now = time.time()
				if current:
					current = (now - t0, current[1], now - current[2],)
				else:
					current = (now - t0, self.job_method, 0,)
				if self.verbose == 'dots':
					if waited % 60 == 0:
						sys.stdout.write('[%d]\n ' % (now - t0,))
					else:
						sys.stdout.write('.')
				elif self.verbose == 'log':
					if waited % 60 == 0:
						print('%d seconds, still waiting for %s (%d seconds)' % current)
				else:
					current_display = (
						fmttime(current[0], True),
						current[1],
						fmttime(current[2], True),
					)
					sys.stdout.write('\r\033[K           %s %s %s' % current_display)
			idle, status_stacks, current, last_time = self._server_idle(1)
		if self.verbose == 'dots':
			print('(%d)]' % (last_time,))
		else:
			print('\r\033[K              %s' % (fmttime(last_time),))

	def jobid(self, method):
		"""
		Return jobid of "method"
		"""
		if 'jobs' in self.job_retur:
			return self.job_retur.jobs[method].link

	def dump_history(self):
		return self.history

	def _server_idle(self, timeout=0, ignore_errors=False):
		"""ask server if it is idle, return (idle, status_stacks)"""
		path = ['status']
		if self.verbose:
			path.append('full')
		path.append('?subjob_cookie=%s&timeout=%d' % (self.subjob_cookie or '', timeout,))
		resp = self._url_json(*path)
		if 'last_error' in resp and not ignore_errors:
			print("\nFailed to build jobs:", file=sys.stderr)
			for jobid, method, status in resp.last_error:
				e = JobError(jobid, method, status)
				print(e.format_msg(), file=sys.stderr)
			raise e
		return resp.idle, resp.get('status_stacks'), resp.get('current'), resp.get('last_time')

	def _server_submit(self, json):
		# submit json to server
		postdata = urlencode({'json': setupfile.encode_setup(json)})
		res = self._url_json('submit', data=postdata)
		if 'error' in res:
			raise ServerError('Submit failed: ' + res.error)
		if 'why_build' not in res:
			if not self.subjob_cookie:
				self._printlist(res.jobs)
			self.validate_response(res.jobs)
		return res

	def _printlist(self, returndict):
		# print (return list) in neat format
		for method, item in sorted(returndict.items(), key=lambda x: x[1].link):
			if item.make == True:
				make_msg = 'MAKE'
			else:
				make_msg = item.make or 'link'
			print('        -  %44s' % method.ljust(44), end=' ')
			print(' %s' % (make_msg,), end=' ')
			if self.print_full_jobpath:
				print(' %s' % Job(item.link).path, end=' ')
			else:
				print(' %s' % item.link, end=' ')
			if item.make != True and 'total_time' in item:
				print(' %s' % fmttime(item.total_time), end=' ')
			print()

	def method_info(self, method):
		return self._url_json('method_info', method)

	def methods_info(self):
		return self._url_json('methods')

	def update_methods(self):
		resp = self._url_get('update_methods')
		self.update_method_info()
		return resp

	def update_method_info(self):
		self._method_info = self.methods_info()

	def list_workdirs(self):
		return self._url_json('list_workdirs')

	def call_method(self, method, options={}, datasets={}, jobs={}, record_in=None, record_as=None, why_build=False, caption=None, workdir=None, **kw):
		if method not in self._method_info:
			raise Exception('Unknown method %s' % (method,))
		info = self._method_info[method]
		params = dict(options=dict(options), datasets=dict(datasets), jobs=dict(jobs))
		argmap = defaultdict(list)
		for thing in ('options', 'datasets', 'jobs'):
			for n in info[thing]:
				argmap[n].append(thing)
		for k, v in kw.items():
			if k not in argmap:
				raise Exception('Keyword %s not in options/datasets/jobs for method %s' % (k, method,))
			if len(argmap[k]) != 1:
				raise Exception('Keyword %s has several targets on method %s: %r' % (k, method, argmap[k],))
			params[argmap[k][0]][k] = v
		jid, res = self._submit(method, caption=caption, why_build=why_build, workdir=workdir, **params)
		if why_build: # specified by caller
			return res.why_build
		if 'why_build' in res: # done by server anyway (because --flags why_build)
			print("Would have built from:")
			print("======================")
			print(setupfile.encode_setup(self.history[-1][0], as_str=True))
			print("Could have avoided build if:")
			print("============================")
			print(json_encode(res.why_build, as_str=True))
			print()
			from inspect import stack
			stk = stack()[2]
			print("Called from %s line %d" % (stk[1], stk[2],))
			exit()
		jid = Job(jid, record_as or method)
		self.record[record_in].append(jid)
		return jid


def fmttime(t, short=False):
	if short:
		units = ['h', 'm', 's']
		fmts = ['%.2f', '%.1f', '%.0f']
	else:
		units = ['hours', 'minutes', 'seconds']
		fmts = ['%.2f ', '%.1f ', '%.1f ']
	unit = units.pop()
	fmt = fmts.pop()
	while t > 60 * 3 and units:
		unit = units.pop()
		fmt = fmts.pop()
		t /= 60
	return fmt % (t,) + unit


class JobList(_ListTypePreserver):
	"""
	A list of Jobs with some convenience methods.
	.find(method) a new JobList with only jobs with that method in it.
	.get(method, default=None) latest Job with that method.
	[method] Same as .get but error if no job with that method is in the list.
	.as_tuples The same list but as (method, jid) tuples.
	.pretty a pretty-printed version (string).
	"""

	def __getitem__(self, item):
		if isinstance(item, slice):
			return self.__class__(list.__getitem__(self, item))
		elif isinstance(item, str_types):
			return self.find(item)[-1] # last matching or IndexError
		else:
			return list.__getitem__(self, item)

	@property
	def pretty(self):
		"""Formated for printing"""
		if not self: return 'JobList([])'
		template = '   [%%3d] %%%ds : %%s' % (max(len(i.method) for i in self),)
		return 'JobList(\n' + \
			'\n'.join(template % (i, j.method, j) for i, j in enumerate(self)) + \
			'\n)'

	@property
	def as_tuples(self):
		return [(e.method, e) for e in self]

	def find(self, method):
		"""Matching elements returned as new Joblist."""
		return self.__class__(e for e in self if e.method == method)

	def get(self, item, default=None):
		try:
			return self[item]
		except IndexError:
			return default

	@property
	def exectime(self):
		total = 0
		seen = set()
		per_method = defaultdict(int)
		for jid in self:
			if jid not in seen:
				seen.add(jid)
				t = jid.post.exectime.total
				total += t
				per_method[jid.method] += t
		return total, per_method

	def print_exectimes(self, verbose=True):
		total, per_method = self.exectime
		if verbose and per_method:
			print("Time per method:")
			tmpl = "   %%-%ds  %%s  (%%d%%%%)" % (max(len(method) for method in per_method),)
			total_time = sum(per_method.values())
			for method, t in sorted(per_method.items(), key=itemgetter(1), reverse=True):
				print(tmpl % (method, fmttime(t), round(100 * t / total_time) if total_time else 0.0))
		print("Total time", fmttime(total))


class UrdResponse(dict):
	def __new__(cls, d):
		assert cls is UrdResponse, "Always make these through UrdResponse"
		obj = dict.__new__(UrdResponse if d else EmptyUrdResponse)
		return obj

	def __init__(self, d):
		d = dict(d or ())
		d.setdefault('caption', '')
		d.setdefault('timestamp', '0')
		d.setdefault('joblist', JobList())
		d.setdefault('deps', {})
		dict.__init__(self, d)

	__setitem__ = dict.__setitem__
	__delattr__ = dict.__delitem__
	def __getattr__(self, name):
		if name.startswith('_') or name not in self:
			raise AttributeError(name)
		return self[name]

	@property
	def as_dep(self):
		return DotDict(timestamp=self.timestamp, joblist=self.joblist.as_tuples, caption=self.caption)

class EmptyUrdResponse(UrdResponse):
	# so you can do "if urd.latest('foo'):" and similar.
	# python2 version
	def __nonzero__(self):
		return False
	# python3 version
	def __bool__(self):
		return False

def _urd_typeify(d):
	if isinstance(d, str):
		d = json.loads(d)
		if not d or isinstance(d, unicode):
			return d
	res = DotDict()
	for k, v in d.items():
		if k == 'joblist':
			v = JobList(Job(e[1], e[0]) for e in v)
		elif isinstance(v, dict):
			v = _urd_typeify(v)
		res[k] = v
	return res

def _tsfix(ts):
	if ts is None:
		return None
	errmsg = 'Specify timestamps as strings, ints, datetimes or (timestamp, integer), not %r' % (ts,)
	if isinstance(ts, (tuple, list,)):
		assert len(ts) == 2, errmsg
		ts, integer = ts
		assert isinstance(integer, int), errmsg
	else:
		integer = None
	if isinstance(ts, (int, date,)):
		ts = str(ts)
	assert isinstance(ts, str_types), errmsg
	assert ts, errmsg
	if integer is None:
		return ts
	else:
		return '%s+%d' % (ts, integer,)

class Urd(object):
	def __init__(self, a, info, user, password, horizon=None):
		self._a = a
		if info.urd:
			assert '://' in str(info.urd), 'Bad urd URL: %s' % (info.urd,)
		self._url = info.urd or ''
		self._user = user
		self._current = None
		self.info = info
		self.flags = set(a.flags)
		self.horizon = horizon
		self.joblist = a.jobs
		self.workdir = None
		auth = '%s:%s' % (user, password,)
		if PY3:
			auth = b64encode(auth.encode('utf-8')).decode('ascii')
		else:
			auth = b64encode(auth)
		self._headers = {'Content-Type': 'application/json', 'Authorization': 'Basic ' + auth}
		self._auth_tested = False
		self._warnings = []

	def _path(self, path):
		if '/' not in path:
			path = '%s/%s' % (self._user, path,)
		return path

	def _call(self, url, data=None, fmt=_urd_typeify):
		assert self._url, "No urd configured for this server"
		url = url.replace(' ', '%20')
		if data is not None:
			req = Request(url, json_encode(data), self._headers)
		else:
			req = Request(url)
		tries_left = 3
		while True:
			try:
				r = urlopen(req)
				try:
					code = r.getcode()
					if code == 401:
						raise UrdPermissionError()
					if code == 409:
						raise UrdConflictError()
					d = r.read()
					if PY3:
						d = d.decode('utf-8')
					return fmt(d)
				finally:
					try:
						r.close()
					except Exception:
						pass
			except HTTPError as e:
				# It seems inconsistent if we get HTTPError or not for 4xx codes.
				if e.code == 401:
					raise UrdPermissionError()
				if e.code == 409:
					raise UrdConflictError()
				tries_left -= 1
				if not tries_left:
					raise UrdError('Error %d from urd' % (e.code,))
				print('Error %d from urd, %d tries left' % (e.code, tries_left,), file=sys.stderr)
			except ValueError as e:
				tries_left -= 1
				msg = 'Bad data from urd, %s: %s' % (type(e).__name__, e,)
				if not tries_left:
					raise UrdError(msg)
				print('%s, %d tries left' % (msg, tries_left,), file=sys.stderr)
			except URLError:
				tries_left -= 1
				print('Error contacting urd (%d tries left)' % (tries_left,), file=sys.stderr)
				if not tries_left:
					raise UrdError('Error contacting urd')
			time.sleep(4)

	def _get(self, path, *a):
		assert self._current, "Can't record dependency with nothing running"
		path = self._path(path)
		assert path not in self._deps, 'Duplicate ' + path
		url = '/'.join((self._url, path,) + a)
		res = UrdResponse(self._call(url))
		if res:
			self._deps[path] = res.as_dep
		self._latest_joblist = res.joblist
		return res

	def _latest_str(self):
		if self.horizon:
			return '<=' + self.horizon
		else:
			return 'latest'

	def get(self, path, timestamp):
		return self._get(path, _tsfix(timestamp))

	def latest(self, path):
		return self.get(path, self._latest_str())

	def first(self, path):
		return self.get(path, 'first')

	def peek(self, path, timestamp):
		path = self._path(path)
		url = '/'.join((self._url, path, _tsfix(timestamp),))
		return UrdResponse(self._call(url))

	def peek_latest(self, path):
		return self.peek(path, self._latest_str())

	def peek_first(self, path):
		return self.peek(path, 'first')

	def since(self, path, timestamp):
		path = self._path(path)
		url = '%s/%s/since/%s' % (self._url, path, _tsfix(timestamp),)
		return self._call(url, fmt=json.loads)

	def list(self):
		url = '/'.join((self._url, 'list'))
		return self._call(url, fmt=json.loads)

	def begin(self, path, timestamp=None, caption=None, update=False):
		assert not self._current, 'Tried to begin %s while running %s' % (path, self._current,)
		if not self._auth_tested:
			try:
				self._call('%s/test/%s' % (self._url, self._user,), True)
			except UrdPermissionError:
				raise Exception('Urd says permission denied, did you forget to set URD_AUTH?')
			self._auth_tested = True
		self._current = self._path(path)
		self._current_timestamp = _tsfix(timestamp)
		self._current_caption = caption
		self._update = update
		self._deps = {}
		self._a.clear_record()
		self.joblist = self._a.jobs
		self._latest_joblist = None

	def abort(self):
		self._current = None

	def finish(self, path, timestamp=None, caption=None):
		path = self._path(path)
		assert self._current, 'Tried to finish %s with nothing running' % (path,)
		assert path == self._current, 'Tried to finish %s while running %s' % (path, self._current,)
		user, build = path.split('/')
		self._current = None
		caption = caption or self._current_caption or ''
		if timestamp is None:
			timestamp = self._current_timestamp
		else:
			timestamp = _tsfix(timestamp)
		assert timestamp, 'No timestamp specified in begin or finish for %s' % (path,)
		data = DotDict(
			user=user,
			build=build,
			joblist=self.joblist.as_tuples,
			deps=self._deps,
			caption=caption,
			timestamp=timestamp,
		)
		if self._update:
			data.flags = ['update']
		url = self._url + '/add'
		return self._call(url, data)

	def truncate(self, path, timestamp):
		url = '%s/truncate/%s/%s' % (self._url, self._path(path), _tsfix(timestamp),)
		return self._call(url, '')

	def set_workdir(self, workdir):
		"""Build jobs in this workdir, None to restore default"""
		self.workdir = workdir

	def build(self, method, options={}, datasets={}, jobs={}, name=None, caption=None, why_build=False, workdir=None, **kw):
		return self._a.call_method(method, options=options, datasets=datasets, jobs=jobs, record_as=name, caption=caption, why_build=why_build, workdir=workdir or self.workdir, **kw)

	def build_chained(self, method, options={}, datasets={}, jobs={}, name=None, caption=None, why_build=False, workdir=None, **kw):
		assert 'previous' not in set(datasets) | set(jobs) | set(kw), "Don't specify previous to build_chained"
		assert name, "build_chained must have 'name'"
		assert self._latest_joblist is not None, "Can't build_chained without a dependency to chain from"
		kw = dict(kw)
		kw['previous'] = self._latest_joblist.get(name)
		return self.build(method, options, datasets, jobs, name, caption, why_build, workdir, **kw)

	def warn(self, line=''):
		"""Add a warning message to be displayed at the end of the build"""
		self._warnings.extend(l.rstrip() for l in line.expandtabs().split('\n'))

	def _show_warnings(self):
		if self._warnings:
			from itertools import chain
			from accelerator.compat import terminal_size
			max_width = max(34, terminal_size().columns - 6)
			def reflow(line):
				indent = ''
				for c in line:
					if c.isspace():
						indent += c
					else:
						break
				width = max(max_width - len(indent), 25)
				current = ''
				between = ''
				for word in line[len(indent):].split(' '):
					if len(current + word) >= width:
						if len(word) > width / 2 and word.startswith('"/') and (word.endswith('"') or word.endswith('",')):
							for pe in word.split('/'):
								if len(current + pe) >= width:
									if current:
										yield indent + current
										current = ('/' if between == '/' else '') + pe
									else:
										yield indent + between + pe
								else:
									current += between + pe
								between = '/'
						else:
							if current:
								yield indent + current
								current = word
							else:
								yield indent + word
					else:
						current += between + word
					between = ' '
				if current:
					yield indent + current
			warnings = list(chain.from_iterable(reflow(w) for w in self._warnings))
			print()
			width = max(len(line) for line in warnings)
			print('\x1b[35m' + ('#' * (width + 6)) + '\x1b[m')
			for line in warnings:
				print('\x1b[35m##\x1b[m', line.ljust(width), '\x1b[35m##\x1b[m')
			print('\x1b[35m' + ('#' * (width + 6)) + '\x1b[m')
			self._warnings = []


def find_automata(a, package, script):
	all_packages = sorted(a.config()['method_directories'])
	if package:
		if package in all_packages:
			package = [package]
		else:
			for cand in all_packages:
				if cand.endswith('.' + package):
					package = [cand]
					break
			else:
				raise Exception('No method directory found for %r in %r' % (package, all_packages))
	else:
		package = all_packages
	if not script.startswith('build'):
		script = 'build_' + script
	for p in package:
		module_name = p + '.' + script
		try:
			module_ref = import_module(module_name)
			print(module_name)
			return module_ref
		except ImportError as e:
			if PY3:
				if not e.msg[:-1].endswith(script):
					raise
			else:
				if not e.message.endswith(script):
					raise
	raise Exception('No build script "%s" found in {%s}' % (script, ', '.join(package)))

def run_automata(options, cfg):
	g.running = 'build'
	a = Automata(cfg.url, verbose=options.verbose, flags=options.flags.split(','), infoprints=True, print_full_jobpath=options.fullpath)

	if options.abort:
		a.abort()
		return

	try:
		a.wait(ignore_old_errors=not options.just_wait)
	except JobError:
		# An error occured in a job we didn't start, which is not our problem.
		pass

	if options.just_wait:
		return

	module_ref = find_automata(a, options.package, options.script)

	assert getarglist(module_ref.main) == ['urd'], "Only urd-enabled automatas are supported"
	if 'URD_AUTH' in os.environ:
		user, password = os.environ['URD_AUTH'].split(':', 1)
	else:
		user, password = os.environ['USER'], ''
	info = a.info()
	urd = Urd(a, info, user, password, options.horizon)
	if options.quick:
		a.update_method_info()
	else:
		a.update_methods()
	module_ref.main(urd)
	urd._show_warnings()


def main(argv, cfg):
	parser = ArgumentParser(
		prog=argv.pop(0),
		usage="%(prog)s [options] [script]",
		formatter_class=RawTextHelpFormatter,
	)
	parser.add_argument('-f', '--flags',    default='',          help="comma separated list of flags", )
	parser.add_argument('-A', '--abort',    action='store_true', help="abort (fail) currently running job(s)", )
	parser.add_argument('-q', '--quick',    action='store_true', help="skip method updates and checking workdirs for new jobs", )
	parser.add_argument('-w', '--just_wait',action='store_true', help="just wait for running job, don't run any build script", )
	parser.add_argument('-F', '--fullpath', action='store_true', help="print full path to jobdirs")
	parser.add_argument('--verbose',        default='status',    help="verbosity style {no, status, dots, log}")
	parser.add_argument('--quiet',          action='store_true', help="same as --verbose=no")
	parser.add_argument('--horizon',        default=None,        help="time horizon - dates after this are not visible in\nurd.latest")
	parser.add_argument('script',           default='build'   ,  help="build script to run. default \"build\".\nsearches under all method directories in alphabetical\norder if it does not contain a dot.\nprefixes build_ to last element unless specified.\npackage name suffixes are ok.\nso for example \"test_methods.tests\" expands to\n\"accelerator.test_methods.build_tests\".", nargs='?')

	options = parser.parse_args(argv)

	if '.' in options.script:
		options.package, options.script = options.script.rsplit('.', 1)
	else:
		options.package = None

	options.verbose = {'no': False, 'status': True, 'dots': 'dots', 'log': 'log'}[options.verbose]
	if options.quiet: options.verbose = False

	try:
		run_automata(options, cfg)
		return 0
	except (JobError, ServerError):
		# If it's a JobError we don't care about the local traceback,
		# we want to see the job traceback, and maybe know what line
		# we built the job on.
		# If it's a ServerError we just want the line and message.
		print_minimal_traceback()
	return 1


def print_minimal_traceback():
	build_fn = __file__
	if build_fn[-4:] in ('.pyc', '.pyo',):
		# stupid python2
		build_fn = build_fn[:-1]
	blacklist_fns = {build_fn}
	_, e, tb = sys.exc_info()
	last_interesting = tb
	while tb is not None:
		code = tb.tb_frame.f_code
		if code.co_filename not in blacklist_fns:
			last_interesting = tb
		tb = tb.tb_next
	lineno = last_interesting.tb_lineno
	filename = last_interesting.tb_frame.f_code.co_filename
	if isinstance(e, JobError):
		print("Failed to build job %s on %s line %d" % (e.jobid, filename, lineno,))
	else:
		print("Server returned error on %s line %d:\n%s" % (filename, lineno, e.args[0]))



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify that the output from methods is captured correctly for all valid
combinations of prepare/analysis/synthesis, both in OUTPUT dir and in
status stacks.
'''

from random import randint
import sys
import os

from accelerator import subjobs
from accelerator.build import Automata
from accelerator import g

def test(params, p=False, a=False, s=False):
	prefix = "A bit of text."
	opts = {'prefix': prefix}
	name = 'test_output_'
	cookie = randint(10000, 99999)
	if p:
		name += 'p'
		opts['p'] = "Some words\nfrom prepare\nwith %d in them." % (cookie,)
	if a:
		name += 'a'
		opts['a'] = "A few words\nfrom analysis(%%d)\nwith the cookie %d in them." % (cookie,)
	if s:
		name += 's'
		opts['s'] = "Words\nfrom synthesis\ncookie is %d." % (cookie,)
	jid = subjobs.build(name, options=opts)
	d = jid.filename('OUTPUT/')
	chked = set()
	all = []
	def chk(part):
		output = jid.output(part)
		if isinstance(part, int):
			data = opts['a'] % (part,)
			part = str(part)
		else:
			data = opts[part[0]]
		chked.add(part)
		with open(d +  part, 'r') as fh:
			got = fh.read()
		want = prefix + '\n' + data + '\n'
		assert got == prefix + '\n' + data + '\n', "%s produced %r in %s, expected %r" % (jid, got, part, want,)
		assert output == got, 'job.output disagrees with manual file reading for %s in %s. %r != %r' % (part, jid, output, got,)
		all.append(got)
	if p:
		chk('prepare')
	if a:
		for sliceno in range(params.slices):
			chk(sliceno)
	if s:
		chk('synthesis')
	unchked = set(os.listdir(d)) - chked
	assert not unchked, "Unexpected OUTPUT files from %s: %r" % (jid, unchked,)
	output = jid.output()
	got = ''.join(all)
	assert output == got, 'job.output disagrees with manual file reading for <all> in %s. %r != %r' % (jid, output, got,)

def synthesis(params):
	test(params, s=True)
	test(params, p=True, s=True)
	test(params, p=True, a=True, s=True)
	test(params, p=True, a=True)
	test(params, a=True, s=True)
	test(params, a=True)


# This is run in all parts in the subjobs.
# The code is here so it's not repeated.

def sub_part(sliceno, opts):
	a = Automata(g.server_url, verbose=True)
	pid = os.getpid()
	def verify(want):
		timeout = 0
		got = None
		for _ in range(25):
			status_stacks = a._server_idle(timeout)[1]
			for line in status_stacks:
				if line[0] == pid and line[1] < 0:
					# this is our tail
					got = line[2]
					if got == want:
						return
			# it might not have reached the server yet
			timeout += 0.01
		# we've given it 3 seconds, it's not going to happen.
		raise Exception("Wanted to see tail output of %r, but saw %r" % (want, got,))
	print(opts.prefix, file=sys.stderr)
	verify(opts.prefix + '\n')
	if isinstance(sliceno, int):
		msg = opts.a % (sliceno,)
	else:
		msg = opts[sliceno]
	print(msg)
	verify(opts.prefix + '\n' + msg + '\n')



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# Test data for use in dataset testing

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from datetime import date, time, datetime

from accelerator.compat import first_value, num_types

# Constraints on this data:
#    Each tuple must be the same length.
#    The tests will not work with less than len(tuple) slices.
#    Time values may not have a 0 microsecond, because str() will omit that.
#    No tuple should contain duplicate values (but bool has to).
#    All values must be sortable (within the same tuple).
#       Which means no tuple may contain None, even for types that support None.
#    Numeric types other than int64 must have a low-ish first value.
#    Most of the above doesn't apply to json (because it's handled specially.)
# It's supposed to contain all types, but it doesn't really have to.
data = {
	"float64": (1/3, 1e100, -9.0),
	"float32": (100.0, -0.0, 2.0),
	"int64": (9223372036854775807, -9223372036854775807, 100),
	"int32": (-2147483647, 2147483647, -1),
	"bits64": (0, 18446744073709551615, 0x55aa55aa55aa55aa),
	"bits32": (0, 4294967295, 0xaa55aa55),
	"bool": (True, False, True,),
	"datetime": (datetime(1916, 2, 29, 23, 59, 59, 999999), datetime(1916, 2, 29, 23, 59, 59, 999998), datetime(1970, 1, 1, 0, 0, 0, 1)),
	"date": (date(2016, 2, 29), date(2016, 2, 28), date(2017, 6, 27),),
	"time": (time(12, 0, 0, 999999), time(12, 0, 0, 999998), time(0, 1, 2, 3)),
	"bytes": (b"foo", b"bar", b"blutti",),
	"unicode": ("bl\xe5", "bl\xe4", "bla",),
	"ascii": ("foo", "bar", "blutti",),
	# big value - will change if it roundtrips through (any type of) float, semibig to find 32bit issues, and a float.
	"number": (1000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000, 13578058080989382, 1/3),
	"json": ({"a": [1, 2, {"b": {}}]}, None, "bl\xe4"),
}

value_cnt = {len(v) for v in data.values()}
assert len(value_cnt) == 1, "All tuples in data must have the same length."
value_cnt = first_value(value_cnt)

not_none_capable = {"bits64", "bits32",}

columns = {t: t if t in not_none_capable else (t, True) for t in data}

def sort_data_for_slice(sliceno):
	# numeric types use only (modified) v[0], other types cycle through their values.
	# int64 goes down one every other line,
	# all other numeric columns go up sliceno + 1 every line.
	# json starts as 0 (pretending to be a numeric type).
	def add(offset):
		res = []
		for k, v in sorted(data.items()):
			if k == "json":
				v = [0]
			if offset == 42 and k.startswith('float'):
				# Should sort last.
				v = float('NaN')
			elif isinstance(v[0], num_types) and k != "bool":
				v = v[0]
				if k == "int64":
					v -= offset // 2
				else:
					v += offset * (sliceno + 1)
			else:
				v = v[offset % len(v)]
			res.append(v)
		return tuple(res)
	for offset in range(sliceno + 2):
		yield add(offset)
	# Add some Nones in the middle
	# (Sorts first except for date/times where they sort last.)
	res = []
	for k in sorted(data):
		if k in not_none_capable:
			res.append(data[k][0])
		elif k == 'json':
			res.append(0)
		else:
			res.append(None)
	yield tuple(res)
	for offset in range(sliceno, 128):
		yield add(offset)
